
package no.rwr.xutil;

import no.rwr.x.*;
import java.awt.Rectangle;


/** Canvas is the non-bare-bones version of an X11 Window. Canvas does
lazy evaluation, event handling and other neat tricks, while Window
simply wraps a window XID. */

/* Here might also goes extra stuff like such as DND, etc. */

/*
  STATES:

  !valid
  valid !mapped
  valid mapped
  


   ,-------. [valid, visible, parent]     ,------.
  |         |--------------------------->|        |
  |         |                  doMap()   |        |
  | !mapped |                            | mapped |
  |         |                            |        |
  |         |<---------------------------|        |
   `-------             [valid, !parent] `------'
                         [valid, !visble]

  checkMapped() performs transition test.

  transitions into and within mapped result in whenMapped being called.

*/


public class Canvas {

    // Remember: do not access window directly. Always use getWindow()
    private Window window;

    Window parent;

    WindowAttributes attributes;
    Rectangle bounds;
    boolean positionChanged;
    boolean sizeChanged;;

    boolean valid = true;

    /** Don't consider <code>valid</code> a state, but rather an
	prerequisite for transit between states. */
    boolean visible = true;

    // intended mapped state
    boolean mapped = false;

    // flushed mapped state
    boolean windowMapped = false;

    public Canvas() {
    }

    public void setParent(Canvas parent) {
	setParent(parent.getWindow());
    }

    public final void setParent(Window parent) {
	this.parent = parent;

	// valid, visible, parent ---> mapped
	if (valid) whenValid();
    }

    public final void setAttributes(WindowAttributes attributes) {
	this.attributes = attributes;
	checkAttributes();
    }

    private final void checkAttributes() {
	if (!(valid && mapped)) return;
	if (attributes == null) return;

	// Remember: do not access window directly. Always use getWindow()
	attributes.apply(getWindow());
	attributes = null;
    }

    public final Window getWindow() {
	if (window == null) {
	    if (attributes == null) attributes = new WindowAttributes();
	    if (bounds == null) bounds = new Rectangle(0, 0, 200, 200);

	    // any default event mask?
	    // attributes.setEventMask(WindowAttributes.MASK_EXPOSURE);

	    window = new CanvasWindow(this, parent, bounds, attributes);
	    sizeChanged = false;
	    positionChanged = false;
	}
	
	return window;
    }

    /* Precondition: valid == true. Called when valid to flush pending
       changes. Call this visibility or parent changes. */
    public void whenValid() {

	// valid, visible, parent ---> mapped
	mapped = visible && (parent != null);
	
	checkMapped();
    }

    public final void setValid(boolean validness) {
	if (valid == validness) return;
	valid = validness;
	if (valid) whenValid();
    }

    public final void setVisible(boolean visible) {
	if (this.visible == visible) return;
	this.visible = visible;
	
	// valid, visible, parent ---> mapped
	if (valid) whenValid();
    }

    private final void checkMapped() {
	// valid, visible, parent ---> mapped
	if (!valid) return; // ignore for now
	
	if (mapped) {
	    whenMapped();
	    if (!windowMapped) {
		getWindow().map();
		windowMapped = true;
	    }
	} else {
	    if (windowMapped) {
		throw new RuntimeException("unmap not implemented");
	    }
	}
    }
	
	
    /* Precondition: visible && (parent != null).  Called to flush
       changes when mapped. */
    protected void whenMapped() {
	checkAttributes();
	checkBounds();
    }


    /* I'm not sure checking for unchanged values is really worth
       it. On the one hand, it can help counteract stupidity on higher
       levels, but ideally stupidity shouldn't exist.

       When feeding a local X server, this might actually decrease
       efficiency in normal situations. A benchmark should probably be
       run at some point.  */

    public void setBounds(Rectangle bounds) {
	setPosition(bounds.x, bounds.y);
	setSize(bounds.width, bounds.height);
	checkBounds();
    }

    public void setPosition(int x, int y) {
	if (bounds == null) {
	    bounds = new Rectangle(x, y, 200, 200);
	} else {
	    if ((bounds.x == x) && (bounds.y == y)) return;
	    bounds.x = x;
	    bounds.y = y;
	}
	positionChanged = true;
	checkBounds();
    }

    public void setSize(int width, int height) {
	if (bounds == null) {
	    bounds = new Rectangle(0, 0, width, height);
	} else {
	    if ((bounds.width == width) && (bounds.height == height))
		return;
	    bounds.width = width;
	    bounds.height = height;
	}
	sizeChanged = true;
	checkBounds();
    }

    public void checkBounds() {
	if (!(valid && mapped)) return;

	getWindow(); // make sure a window has been created
	if (sizeChanged) {
	    if (positionChanged) {
		throw new
		    RuntimeException("changing bounds not implemented");
		// positionChanged = false;
	    } else {
		throw new
		    RuntimeException("changing size not implemented");
	    }
	    // sizeChanged = false;
	} else if (positionChanged) {
	    throw new
		RuntimeException("changing position not implemented");
	    // positionChanged = false;
	}
    }
    
    public void processEvent(XEvent event) {

	int type = event.getType();
	//System.out.println("Got event, " + event + ", type=" + type);
	switch (type) {
	case XEvent.TYPE_CLIENT_MESSAGE:
	    processClientMessage(new ClientMessageEvent(event));
	    return;
	}

	if (type == XEvent.TYPE_EXPOSE) {
	    exposed(new XExposeEvent(event));
	} else {

	    System.err.println("Unknown event type: " + type);

	}

    }
	
    protected void exposed(XExposeEvent xee) {
    }

    public void processClientMessage(ClientMessageEvent event) {
	System.err.println("Unknown client message: " + event);
    }

}

